
/*	Example/Tutorial for the zrgUnit test suite.
	
	Author: Markus Broecker	<mbroecker@uni-koblenz.de>

	This is a small example, where a provided Vector3 class will be tested (not completely though). It
	will show you the basic steps that you need to take, to get this show on the road.
*/

/*	First include zrgUnit.h This is all you need. No need for anything else or libraries. Small, clean,
	nice :)
*/
#include "zrgUnit.h"

//	Vector3 is an older Vector 3 class I wrote sometime ago. We'll use it for testing. 
#include "vector3.h"


/*	In this example, we will test the provided Vector 3 class. Because we know what Vector3 will do, we
	can write some tests, to check if the maths functions are implemented correctly. The first thing one
	has to do, is to write a class that inherits from zrgUnit::TestSuite.
*/
class Vector3Test : public zrgUnit::TestSuite
{
	public:
		/* A Test suite always has a name. 
		*/
		Vector3Test() : zrgUnit::TestSuite("Vector3 Tests")
		{
		}
		
	protected:
		/*	The first test will check if we have SIMD capabilities. Enabled with -msse on the gcc compilers.
		*/
		void testSIMD()
		{
			#ifndef __SSE__
				// just bail out, if are not cool
				zrgUnit::fail("SSE not enabled");
			#endif
		}
		
		/*	The next test will check if the overloaded operator == works, as intended. The other tests make
			use of this operator for convenience, so it should work.
		*/
		void testEquality()
		{
			/*	Note that everything is in the namespace zrgUnit. If you are really lazy, just add using namespace
				zrgUnit at the top of your header or source files ... 
			*/
			using namespace zrgUnit;
		
			Vector3 a(1, 2, 3);
			Vector3 b(1, 2, 3);
			
			// check the single values
			assertEqual( a.x, b.x );
			assertEqual( a.y, b.y );
			assertEqual( a.z, b.z);
			
			// check the operator
			assertEqual( a, b );
		}
		
		/*	Now we test the constructors and assignment operators.
		*/
		void testCTOR()
		{
			using namespace zrgUnit;
		
			// simple ctor
			Vector3 a(1, 2, 3);
			assertNotNULL( &a );
			assertEqual(a.x, 1.f);
			assertEqual(a.y, 2.f);
			assertEqual(a.z, 3.f);
			
			// copy ctor
			Vector3 b(a);
			assertEqual(a, b);
			
			// assignment
			Vector3 c = a;
			assertEqual(c, a);
		}

		/* Test some scalar math functions
		*/
		void testScalarMath()
		{
			using namespace zrgUnit;
		
			// operator *
			Vector3 a(1, 2, 3);
			assertEqual(a*2, Vector3(2, 4, 6, 2));
			
			// operator *=, / and /=
			a *= 2;
			assertEqual(a, Vector3(2, 4, 6, 2));
			assertEqual(a/2, Vector3(1, 2, 3));
			a /= 2;
			assertEqual(a, Vector3(1, 2, 3));
		}
		
		/* Test some vector math
		*/
		void testVectorMath()
		{
			using namespace zrgUnit;
			
			Vector3 a(1, 0, 0, 0);
			Vector3 b(0, 1, 0, 0);
			
			// addition and subtraction
			assertEqual(a+b, Vector3(1,1,0));
			assertEqual(a-b, Vector3(1, -1, 0));
					
			// check the length
			Vector3 c(1, 0, 0);
			assertEqual(c.length(), 1.f);
			
			Vector3 d(2, 52, -3);
			d.normalize();
			//std::clog << d.length();
			// this will fail because of numeric errors. Uncomment line above
			assertEqual(d.length(), 1.f);
			
			// cross product
			Vector3 n = cross( a, b );
			assertEqual(n, Vector3(0, 0, 1) );
			
		}
	
		/*	Now the interesint part. We tell the test suite where it can find the test cases.
			In JUnit, it's quite easy, using reflection. So here in C++, we have to deal with
			function pointers, which are very ugly. Fortunately there are some convenience
			functions provided to make life easier. 
			A test case always gets a unique name to differentiate it from the others.
		*/
		void buildTestCases()
		{
			/*	Here we add the SIMD test. The first parameter is a string with the name of the
				test case. The second is the memory address of the method of this class
					&Vector3Test::testSIMD
				The & operator gets the memory location, Vector3Test is the name of our test suite
				and testSIMD is the name of the method we want as a test case. This method _must not_
				take any arguments and must return void. At last, we typecast this to a (TestCase)
				and we're done. TestCase is a typedef in TestSuite, our parent class. We have to
				typecase, so the compiler will shut up. And, we're done.
				
				Please note that the tests are not run in this particular order, but in an alphabetical
				order. But on the other hand, no test should depend on another one.
			*/					
			addTest("SIMD Avail?", (TestCase)&Vector3Test::testSIMD );
			addTest("Operator ==", (TestCase)&Vector3Test::testEquality );
			addTest("Conctructor", (TestCase)&Vector3Test::testCTOR );
			addTest("Scalar Math", (TestCase)&Vector3Test::testScalarMath );
			addTest("Vector Math", (TestCase)&Vector3Test::testVectorMath );
		}
		
		/*	setUp and tearDown are called fixtures. They are used to create a test environment for the test
			cases. They are run every time a test case is run: Setup before the test, tearDown after the
			test finished.
		*/
		void setUp()
		{
			// doNothing();
		}
		
		/*	Although not needed in this example, it just shows that you are able to override them.
		*/
		void tearDown()
		{
			// doNothing();
		}
};


int main (int argc, char * const argv[]) 
{
	// Create an instance of out test suite
	Vector3Test suite;
	
	// And create a test runner. This class runs the test cases in a suite and prints info 
	zrgUnit::TestRunner runner;
	
	// Run the tests
	runner.run( suite );
	
		
	return 0;
}
