/*
 *      vector3.h
 *
 *      Copyright 2007 Markus Broecker <bluekahuna@users.sourceforge.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef Vector3_H_INCLUDED
#define Vector3_H_INCLUDED

#include <cmath>
#include <iostream>

#ifdef __SSE__
#include <xmmintrin.h>
#endif



/*
#ifdef __SSE2__
__declspec(align(16))
#endif
*/
/** \brief A 3-dimensional vector with single precision (float). If the
	compile flag SSE is enabled (for example in march=pentium4, it will
	use SSE	SIMD instructions to speed things up.
	Contains assorted functions.
	@author M. Broecker
	@version 2.0
**/
class Vector3
{
	public:
		/// Ctor
		/** Constructs an untinitialized vector with undefined attributes.
		 */
		Vector3() {}

		/// Ctor
		/**	Copy constructor.
			@param cp Original vector.
		*/
		Vector3(const Vector3& cp);

		/// Ctor
		/** Four floats make up a 3d vector. The default w values is 1.0.
			@param x x component
			@param y y component
			@param z z component
			@param w homogenous coordinate
		*/
		Vector3(float x, float y, float z, float w = 1.0);

		#ifdef __SSE__
		/// C'Tor
		/**	Create a new Vector from an existing SIMD vector.
			@param vec existing 128bit SSE SIMD vector
		*/
		Vector3(__m128 vec) : f(vec)
		{
		}

		#endif

		/// Resets vector
		/**	Sets the vector to the Null vector (0.0, 0.0, 0.0, 1.0)
		*/
		inline void zero()
		{
			#ifdef __SSE__
			f = _mm_setzero_ps();
			#else
			x = y = z = 0.0;
			#endif
			w =  1.0;
		}

		/// Magnitude of the vector
		/**	Returns the length = magnitude of this vector. Because this
			involves a square root, this operation is costly and instead
			length2 should be called.
			@return length
		*/
		inline float length() const
		{
			return sqrtf( x*x+y*y+z*z );
		}

		/// Squared Magnitude of the vector
		/**	Returns the squared magnitude of the vector.
			@return squared length
		*/
		inline float length2() const
		{
			return x*x + y*y + z*z;
		}

		/// Normalize the vector
		/**	Normalizes the vector. After this operation it will have
			the same direction as before, but a length of 1.0
		*/
		void normalize();

		/// Inverts the vector
		/**	Flips the vector by multiplying it with -1.0
		*/
		inline void invert()
		{
			#ifdef __SSE__
				f = _mm_mul_ps( f, _mm_set_ps1(-1.0f));
			#else
				x *= -1.0f;
				y *= -1.0f;
				z *= -1.0f;
				w *= -1.0f;
			#endif
		}

		/// Equality
		/**	Tests for equality of 2 vectors. Note that also the w-values
			are compared.
			@param v other vector
			@return true if they are the same, false otherwise
		*/
		inline bool operator == (const Vector3& v) const
		{
			return (x==v.x && y==v.y && z==v.z);
		}

		/// Addition
		/**	Adds two vectors and returns the result
			@param v second vector
			@return result (this +	v)
		*/
		inline Vector3 operator + (const Vector3& v) const
		{
			#ifdef __SSE__
				return Vector3( _mm_add_ps(f, v.f) );
			#else
				return Vector3( x+v.x, y+v.y, z+v.z );
			#endif
		}

		/// Subtraction
		/**	Subtracts the second vector from the first and returns the
			result.
			@param v second vector
			@return result (this - v)
		*/
		inline Vector3 operator - (const Vector3& v) const
		{
			#ifdef __SSE__
				return Vector3( _mm_sub_ps(f, v.f) );
			#else
				return Vector3( x-v.x, y-v.y, z-v.z );
			#endif
		}

		/// Scalar multiplication
		/**	Multiplies the components of the vector with a scalar.
			@param s scalar value
			@return result (this * s)
		*/
		inline Vector3 operator * (float s) const
		{
			#ifdef __SSE__
				return Vector3( _mm_mul_ps(f, _mm_set_ps1(s) ) );
			#else
				return Vector3( x*s, y*s, z*s, w*s );
			#endif
		}

		/// Scalar division
		/**	Divides the components of the vector by an scalar.
			@param s scalar value
			@return reuslt (this / a)
		**/
		Vector3 operator / (float s) const
		{
			#ifdef __SSE__
				return Vector3( _mm_div_ps(f, _mm_set_ps1(s) ) );
			#else
				return Vector3( x/s, y/s, z/s );
			#endif
		}

		/// Addition
		/**	Inline addition.
			@param v second vector
			@return modified vector
		*/
		const Vector3& operator += (const Vector3& v)
		{
			#ifdef __SSE__
				f = _mm_add_ps(f, v.f);
			#else
				x += v.x;
				y += v.y;
				z += v.z;
			#endif
			return *this;
		}

		/// Substraction
		/** Inline substraction
			@param v second vector
			@return modified vector
		*/
		const Vector3& operator -= (const Vector3& v)
		{
			#ifdef __SSE__
				f = _mm_sub_ps(f, v.f);
			#else
				x += v.x;
				y += v.y;
				z += v.z;
			#endif
			return *this;
		}

		/// Multiplication
		/**	inline multiplication
			@param scalar scalar value
			@return modified vector
		*/
		const Vector3& operator *= (float scalar)
		{
			#ifdef __SSE__
				f = _mm_mul_ps(f, _mm_set_ps1(scalar) );
			#else
				x *= scalar;
				y *= scalar;
				z *= scalar;
			#endif
			return *this;
		}

		/// Division
		/**	Inline scalar component division
			@param scalar scalar value
			@return modified vector
		*/
		const Vector3& operator /= (float scalar)
		{
			#ifdef __SSE__
				f = _mm_div_ps(f, _mm_set_ps1(scalar) );
			#else
				x *= scalar;
				y *= scalar;
				z *= scalar;
			#endif
			return *this;
		}

		union
		{
			struct
			{
				/// x-component
				float x;
				/// y-component
				float y;
				/// z-component
				float z;
				/// homogenous coordinate
				float w;		// padding and matrix mult.
			};
			/// combined coordinates
			float coord[4];
			#ifdef __SSE__
			/// 128bit SIMD SSE register (holds 4 floats)
			__m128	f;
			#endif
		};

};

/// Cross product
/**	Calculates the cross product btw two vectors.
	@param lhs Left hand side vector
	@param rhs Right hand side vector
	@return Resulting perpendicular vector.
*/
Vector3 cross(const Vector3& lhs, const Vector3& rhs);


/// Dot product
/**	Calculates the dot product (inner product)
	@param lhs First vector
	@param rhs Second vector
	@return The dot product
*/
static inline float dot(const Vector3& lhs, const Vector3& rhs)
{
	return lhs.x * rhs.x + lhs.y * rhs.y + lhs.z * rhs.z;
}

/// ouput
/**	Writes the vector to a std::ostream in the form (x, y, z). You don't
	use this function directyl, rather write
	\begincode
		Vector3 a(1, 0, 1);

		std::cout << a << std<<endl;
	\endcode
	@param os outstream to write to
	@param v Vector object to write into the stream
	@return modified ostream
*/
static inline std::ostream& operator << (std::ostream& os, const Vector3& v)
{
	return os << "(" << v.x << "," << v.y << "," << v.z << ")";
}

/** Interpolates linear between two vectors. Note that this is a component
 * 	wise interpolation, so the length of the vector _will_ change.
 * 	@param v1 first vector
 *  @param v2 second vector
 *  @param delta position btw. [0..1]
 * 	@return linear interpolated vector.
 */
static inline const Vector3 lerp(const Vector3& v1, const Vector3& v2, double delta)
{
	Vector3 result;
	#ifdef	__SSE__
	result.f = _mm_add_ps(v1.f, _mm_mul_ps( _mm_sub_ps(v2.f, v1.f),_mm_set_ps1(delta)));
	#else
	result = v1 + (v2 - v1) * delta;
	#endif
	return result;
}


#endif // Vector3_H_INCLUDED
