/*
 *  zrgUnit - A lightweight unit testing framework for C++
 *
 * Copyright (c) 2008, Markus Broecker <mbroecker@uni-koblenz.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of Koblenz nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MARKUS BROECKER ''AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MARKUS BROECKER BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef ZRGUNIT_INCLUDED
#define ZRGUNIT_INCLUDED

#include <iostream>
#include <string>
#include <map>

namespace zrgUnit
{

/// The 'library's' version
static const unsigned int VERSION = 3;

/** A kind of exception thrown when an assertion fails.

	@author mbroecker
	@version 1.0
*/
class FailedAssertion
{
	public:
		/// Constructor
		FailedAssertion(const std::string& reason) : mReason(reason)
		{
		}
		
		/// Gets the reason for the assertion
		inline const std::string& getReason() const
		{
			return mReason;
		}
		
	private:
		std::string		mReason;
};

/**	A TestSuite groups test cases. It always has a name, by which it is identified. It is an abstract class,
	you must derive your own test suites from this class and implement the buildTestCases method, where you'll
	tell the Testsuite which of your functions are test cases.

	@author mbroecker
	@version 1.2
*/
class TestSuite
{
	/// TestRunner actually runs the tests
	friend class TestRunner;

	protected:
		/// A member function pointer
		/** TestCase is a function pointer to a method of TestSuite and all of it's derived classes, that take
			no arguments and return nothing.
		*/
		typedef void (TestSuite::*TestCase)();
		
		/// The pool of tests
		typedef std::map<std::string, TestCase> TestList;

	public:
		/// CTor
		/** Every Test case has a name by which it is identified.
			@param name Name of this test case.
		*/
		TestSuite(const std::string& name) : mName(name)
		{
		}

		/// DTor
		virtual ~TestSuite()
		{
		}
		
		/// Gets the name of this test suite.
		inline const std::string& getName() const throw()
		{
			return mName;
		}
		
		/// Gets the number of test cases registered. 
		/** If buildTestCases() was run, this method will return the number of test cases in the pool.
			@return Number of tests.
		*/
		inline const unsigned int getTestCount() const throw ()
		{
			return mTests.size();
		}
		
	protected:		
	
		/// Builds the list of test cases
		/** This method must be implemented in your concrete test suite. It adds test cases to a pool of tests. You should use
			the addTest() method to do that. It's not necessary to call this method manually, as the test runner will call it 
			before commencing the tests.
			
			A test case is a single method in your Testsuite, which takes no parameters and returns void. It can be const and may 
			(and should) throw exceptions, as these are caught by the TestRunner. Valid test cases are for example:
			\code
			void f();
			void g();
			void z() const;
			// you should get the drift by now ...
			\endcode
			Invalid test cases are e.g:
			\code
			int meh();				// integer return value
			void foo(double d);		// no parameters allowed
			int argh[void* ptr];	// not even a valid definition ..
			\endcode
			
			For more info on adding test cases, see addTest()
			
		*/	
		virtual void buildTestCases() = 0;
		
		
		/// \name Fixture
		/// \{
		
		/** This method is called every time before a test is run. It can be used to set up a specific testing
			environment, such as initializing variables, etc.
		*/
		virtual void setUp()
		{
		}
		
		/** This method is called every time after a test is run. It can be used to tear down the testing
			environment, such as cleaning up, releasing memory, etc.
		*/
		virtual void tearDown()
		{
		}
		
		/// \}
		

		/// Adds a test case to the suite
		/** A test case is identified by a name. TestCase is a function pointer to a member function of a derived class 
			that takes no arguments and returns nothing. You should cast your pointer to a TestCase when adding test cases. 
			Getting the pointer of a memeber function is done the following way:
			\code
				addTest( (TestCase)&MyClass:f )
			\endcode
			In this example MyClass is a derived class from TestSuite and has a void f() method. &MyClass:f gets the function pointer
			to this method, and finally it is typecasted to a TestCase function pointer (which is just a typedef for more nasty stuff).
			You don't have to understand or believe this to use this library, trust me ;) 
			A more complete Example:			
			\code
			
			class Foo
			{
				// ...
				
				/// A test case
				void f()
				{
					...
				}
				
				/// Another test case
				void g()
				{
					...
				}
								
				/// Builds the list of test cases
				void buildTestCases()
				{
					// Get the pointer of a function of this class
					addTest("First test", (TestCase)&Foo::f );
					// And don't forget to typecast it to a (TestCase)! 
					addTest("Second test", (TestCase)&Foo:g );
				}
			};
			
			\endcode
		*/	
		void addTest(const std::string& name, TestCase test)
		{
			mTests[name] = test;
		}

	private:
		TestList			mTests;
		std::string			mName;
};

/** A TestRunner runs TestCases and prints the output to the console. You can modify the output of the TestRunner
	by inheriting from it and overwriting the virtual output methods.
	
	@author mbroecker
	@version 1.2
*/
class TestRunner
{
	public:
		/// DTor
		virtual ~TestRunner()
		{
		}


		/// Run this test suite
		/** Runs the test of this suite and calls the appropriate callback functions.
		
			@param suite A test suite
		*/
		void run(TestSuite& suite) throw ()
		{
			testStart( suite.getName() );
		
			suite.buildTestCases();
		
			unsigned int testCounter = 1;
			const unsigned int testCount = suite.mTests.size();

			mSuccess = mFail = mError = 0;

			TestSuite::TestList::iterator it = suite.mTests.begin();
			while( it != suite.mTests.end() )
			{
				
				testRun(testCounter, testCount, it->first);
				
				// place the fixture outside, so it will be always called
				suite.setUp();
				
				try
				{
					(suite.*it->second)();
					testSuccess();
					
					++mSuccess;

				}
				catch (zrgUnit::FailedAssertion fa)
				{
					testFail( fa.getReason() );
					++mFail;
				}  
				catch( std::exception e)
				{
					testError( e.what() );
					++mError;
				}
				catch ( ... )
				{
					testError("unknown error.");
					++mError;
				}
				
				suite.tearDown();
				
				++testCounter;
				++it;
			}
			
			testEnd();
		}
		
		protected:
			unsigned int mSuccess, mFail, mError;
		
		
			/// \name Output callbacks
			/// \{
		
			/// Gets called when the tests start
			virtual void testStart(const std::string& suiteName)
			{
				std::cout << "zrgUnit, Version " << VERSION << " Testrunner." << std::endl;
				std::cout << "Running test suite " << suiteName << std::endl;
				std::cout << "------------------------------------------------------------" << std::endl;
			}
			
			/// Gets called for every single test
			/** @param index The current number of the test
				@param max The total number of tests in this suite.
				@param name The name of the current test case
			*/
			virtual void testRun(unsigned int index, unsigned int max, const std::string& name)
			{
				std::clog << "Test [" << index << "/" << max << "] " << name << " ";
			}
		
			/// Gets called if the test was successful
			virtual void testSuccess()
			{
				std::clog << "succeeded." << std::endl;
			}
			
			/// Gets called when an assertion in the test failed
			virtual void testFail(const std::string& reason)
			{
				std::clog << "failed, reason: " << reason << std::endl;
			}
			
			/// Gets called when an error occured in the test 
			virtual void testError(const std::string& errorText)
			{
				std::clog << "had error(s), text: " << errorText << std::endl;
			}
			
			/// Gets called at the end of the test.
			virtual void testEnd()
			{
				std::cout << "------------------------------------------------------------" << std::endl;
				std::cout << "Success: " << mSuccess << " Fail: " << mFail << " Error:" << mError << std::endl;
			}
			
			/// \}
			
			
		
};


/// \name Assertions
/// \{


/// Assert equality
/** Asserts that two values of the same type are equal. They are, if the operator == is implemented 
	(by default on the basic types like int, float, ...) and it returns true.
	
	@param a first value
	@param b second value
	@throws FailedAssertion if the values are not equal
*/
template<typename T>
inline void assertEqual(const T& a, const T& b) throw (FailedAssertion)
{
	if (!(a == b))
		throw FailedAssertion("Values are not equal!");
}

/// Assert inequality
/** Asserts that two values of the same type are different. They are, if the operator == is implemented 
	(by default on the basic types like int, float, ...) and it returns false.
	
	@param a first value
	@param b second value
	@throws FailedAssertion if the values are equal
*/
template<typename T>
inline void assertNotEqual(const T& a, const T& b) throw (FailedAssertion)
{
	if (a == b)
		throw FailedAssertion("Values are equal!");
}


/// Assert true
/** Asserts that a value is true.
	@param value Value to test
	@throws FailedAssertion if the value is false
*/
inline void assertTrue(bool value) throw (FailedAssertion)
{
	if (!value)
		throw FailedAssertion("Value is not true");
}

/// Assert false
/** Asserts that a value is false.
	@param value Value to test
	@throws FailedAssertion if the value is true
*/
inline void assertFalse(bool value) throw (FailedAssertion)
{
	if (value)
		throw FailedAssertion("Value is not false");
}

/// Assert that a pointer is NULL
/** Asserts that a pointer is not initialized or NULL (0).
	@param ptr Pointer to test
	@throws FailedAssertion if the pointer has been initialized
*/
template<typename T>
inline void assertNULL(const T* ptr) throw (FailedAssertion)
{
	if (ptr != 0)
		throw FailedAssertion("Pointer is not NULL");
}

/// Assert that a pointer is not NULL
/** Asserts that a pointer is initialized and not NULL (0).
	@param ptr Pointer to test
	@throws FailedAssertion if the pointer has not been initialized
*/
template<typename T>
inline void assertNotNULL(const T* ptr) throw (FailedAssertion)
{
	if (ptr == 0)
		throw FailedAssertion("Pointer ist NULL");
}

/// Fails the test
/** This assert will always throw a FailedAssertion with a user specified
	text. It can be used to fail tests deliberately and to exit them early.
	@param failText Reason given to the FailedAssertion
	@throws FailedAssertion with the user specified reason
*/
inline void fail(const std::string& failText = "User specified") throw (FailedAssertion)
{
	throw FailedAssertion(failText);
}

/// \}


}

#endif
